# Copyright 2024 Oreum Industries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# model.describe.py
"""Model Descriptions"""

import re

import arviz as az
import numpy as np
import pandas as pd
import patsy as pat

from ..model_pymc import BasePYMCModel

__all__ = [
    "model_desc",
    "extract_yobs_yhat",
    "describe_dist",
    "get_summary",
    "print_rvs",
]

RSD = 42
rng = np.random.default_rng(seed=RSD)


def model_desc(fml: str) -> str:
    """Convenience: return patsy modeldesc
    NOTE: `.describe()` doesn't return the `1 +` (intercept) term in the
        case that it's present. check and add if needed
    """
    fmls = fml.split(" ~ ")
    add_intercept = False if re.match(r"1 \+", fml) is None else True
    r = pat.ModelDesc.from_formula(fml).describe()
    if len(fmls) == 2:
        rs = r.split(" ~ ")
        if add_intercept:
            r = f"{rs[0]} ~ 1 + {rs[1]}"
    elif len(fmls) == 1:
        if add_intercept:
            r = f"1 + {r[2:]}"
    else:
        raise ValueError("fml must have only a single tilde `~`")
    return f"patsy linear model desc:\n{r}\n"


def extract_yobs_yhat(
    idata: az.InferenceData, obs: str = "y", pred: str = "yhat"
) -> tuple:
    """Convenience: extract y_obs, y_hat from idata
    get yhat in the shape (nsamples, nobs)
    """
    nsamp = np.product(idata.posterior_predictive[pred].shape[:-1])
    yobs = idata.constant_data[obs].values  # (nobs,)
    yhat = idata.posterior_predictive[pred].values.reshape(nsamp, -1)  # (nsamp, nobs)
    return yobs, yhat


def describe_dist(mdl: BasePYMCModel, log: bool = False, inc_summary: bool = False):
    """Convenience: get distribution descriptions from distributions.DistNumpy
    and return for printing or Markdown
    NOTE: consider deprecating
    """
    title = f"{mdl.name}: Natural Distributions"
    dist = mdl.dist_natural
    if log:
        title = f"{mdl.name}: Logged Distributions"
        dist = mdl.dist_log

    if inc_summary:
        return title, {**mdl.notation, **dist, **mdl.conditions, **mdl.summary_stats}
    return title, {**dist}


def get_summary(mdl: BasePYMCModel, rvs: list, group="posterior") -> pd.DataFrame:
    """Convenience fn to get arviz summary of idata posteriors"""
    df = az.summary(mdl.idata, var_names=rvs, group=group)
    return df


def print_rvs(mdl: BasePYMCModel) -> list[str]:
    """Convenience to provide RV strings to display in Notebook
    Use as _ = [display(Markdown(s for s in mt.print_rvs(mdl)))]
    """
    r = []
    for k, rvs in mdl.get_rvs().items():
        if k in ["free", "potentials", "deterministics"]:
            for rv in rvs:
                try:
                    r.append(rv.str_repr(formatting="string", include_params=True))
                except AttributeError:
                    # initially developed as a bit of a hack to deal with
                    # 'TensorVariable' object has no attribute 'str_repr'
                    # in the case of e.g. autoimputed x_mv_unobserved
                    pass
                except AssertionError:
                    pass
    return r


# def print_rvs(rvs: list[pt.TensorVariable]) -> None:
#     """Display rvs to Notebook using latex, post sub underscores
#     Hack to replace print(mdl.model) see https://github.com/pymc-devs/pymc/issues/6869
#     """
#     rx_name_usc = re.compile(r"(?:text\{[^\_\}]+?)(\_+)(?:[^\_\}]+?)(?:(\_+)(?:[^\_\}]+?))*?(?:\})", re.I)
#     for rv in rvs:
#         s = rv.str_repr(formatting='latex', include_params=True)
#         t = rx_name_usc.sub(r"\\_", s)
#         display(Markdown(t))



# Copyright 2024 Oreum Industries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# eda/
"""Various classes & functions for exploratory data analysis"""

from .calc import (
    bootstrap,
    bootstrap_index_only,
    bootstrap_lr,
    calc_geometric_cv,
    calc_location_in_ecdf,
    calc_svd,
    calc_umap,
    fit_and_plot_fn,
    get_gini,
    month_diff,
    tril_nan,
)
from .describe import describe, display_fw, display_ht, get_fts_by_dtype
from .eda_io import FigureIO, output_data_dict
from .plot import (  # plot_umap,; plot_r2_range,; plot_r2_range_pair,
    plot_accuracy,
    plot_binary_performance,
    plot_bool_ct,
    plot_bootstrap_delta_grp,
    plot_bootstrap_grp,
    plot_bootstrap_lr,
    plot_bootstrap_lr_grp,
    plot_cat_ct,
    plot_cdf_ppc_vs_obs,
    plot_coverage,
    plot_date_ct,
    plot_estimate,
    plot_explained_variance,
    plot_f_measure,
    plot_float_dist,
    plot_grp_ct,
    plot_heatmap_corr,
    plot_int_dist,
    plot_joint_numeric,
    plot_kj_summaries_for_linear_model,
    plot_mincovdet,
    plot_rmse_range,
    plot_roc_precrec,
    plot_smrystat,
    plot_smrystat_grp,
    plot_smrystat_grp_year,
)



# Copyright 2024 Oreum Industries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# curate.text_clean.py
"""Text Cleaning"""

import re
import string

import ftfy
import numpy as np

__all__ = ["TextCleaner"]


class TextCleaner:
    """Potentially useful text cleaing - strips lots of crud.
    Originally developed a number of years ago, so may need revision
    """

    def __init__(self):
        """Init with lots of regexes"""
        self.rx_line = re.compile(re.escape("=\n"))  # "=\n"
        self.rx_nbsp = re.compile(r"&nbsp;")  # nbsp
        self.rx_copy = re.compile(r"&copy;")  # nbsp
        self.rx_numbers = re.compile(
            r"(\b[0-9,.\-\\\/]+\b)"
        )  # number blocks, money amounts, years etc
        self.rx_punct = re.compile(
            r"[{}]+".format(re.escape(string.punctuation))
        )  # regular punctuation
        # https://en.wikipedia.org/wiki/Wikipedia:List_of_English_contractions
        self.rx_neg_apostrophe = re.compile(
            r"""\b(ai|are|ca|could|did|do|does|is|had|has|hav|may|might|must|
                    shall|should|was|were|wo|would)(n)(?:\')(t)\b""",
            re.I,
        )
        self.rx_hex = re.compile(r"=[a-f0-9]{2}", re.I)  # stray hexadecimal
        self.rx_arrows = re.compile("(>)+")  # sequences of ">" at start of line
        self.rx_repeatgt3 = re.compile(r"(.)\1{3,}")  # any char repeating more than 3x
        self.rx_htmlcom = re.compile(
            re.escape("<!--") + "(.*?)" + re.escape("-->"), re.DOTALL
        )  # HTML comments (usually embedded CSS)
        self.rx_email = re.compile(
            r"""\b[a-z0-9\'._%+-]+@[a-z0-9.-]+
                                    \.[a-z]{2,4}\b""",
            re.I,
        )  # email addresses
        self.rx_web = re.compile(
            r"""((http[s]?\:\/\/)?(www\.)+
                                    [a-z0-9\-%\/.]+)""",
            re.I,
        )  # web addresses
        self.postcode = re.compile(
            r"""(GIR ?0AA|[A-PR-UWYZ]([0-9]{1,2}|([A-HK-Y][0-9]([0-9ABEHMNPRV-Y])?)
                |[0-9][A-HJKPS-UW])?[0-9][ABD-HJLNP-UW-Z]{2})""",
            re.I,
        )
        self.natins = re.compile(
            r"\s*[a-zA-Z]{2}(?:\s*\d\s*){6}[a-zA-Z]?\s*"
        )  # UK national insurance ID number
        self.phoneno = re.compile(
            r"""\(?(?:(?:0(?:0|11)\)?[\s-]?\(?|\+)44\)?[\s-]?\(?(?:0\)?[\s-]?
                \(?)?|0)(?:\d{2}\)?[\s-]?\d{4}[\s-]?\d{4}|\d{3}\)?[\s-]?\d{3}
                [\s-]?\d{3,4}|\d{4}\)?[\s-]?(?:\d{5}|\d{3}[\s-]?\d{3})|\d{5}\)?
                [\s-]?\d{4,5}|8(?:00[\s-]?11[\s-]?11|45[\s-]?46[\s-]?4\d))
                (?:(?:[\s-]?(?:x|ext\.?\s?|\#)\d+)?)"""
        )  # phone number
        self.nhs = re.compile(r"\d{3}\s?\d{3}\s?\d{4}")  # UK NHS ID number
        # self.rx_nonchar = re.compile(r'[\d{}]+'.format(
        #     re.escape(string.punctuation)))
        self.rx_num_m = re.compile(
            r"^(?P<mill>[0-9]+?(?:[.]+?[0-9]+?)*?)m$", re.I
        )  # (1.4)M
        self.rx_num_k = re.compile(
            r"^(?P<thou>[0-9]+?(?:[.]+?[0-9]+?)*?)k$", re.I
        )  # (400)k
        self.rx_num = re.compile(
            r"^(?P<whol>[0-9]+?)(?P<frac>[.]+?[0-9]+?)*?$", re.I
        )  # (81)(.23)
        self.rx_number_junk = re.compile(r"[#$€£₤¥,;%]")

    def fix_unicode(self, txt: str) -> str:
        """Fix bad unicode / emojis etc and try to remove crud"""

        t = ftfy.fix_text(txt, fix_character_width=False)  # fix encoding
        t = self.rx_hex.sub(r"", t)  # remove hex like '=b7', '=f5' etc

        return t

    def basic_clean(self, txt: str) -> str:
        """Clean up single raw text string where words have single spaces
        Note:
            doesnottokenise
            dOes Not chAnge Casing: allows proper noun removal later
        """
        t = self.rx_line.sub(r"", txt)
        t = self.rx_arrows.sub(r"", t)
        t = self.rx_repeatgt3.sub(r"", t)
        t = self.rx_nbsp.sub(r"", t)
        t = self.rx_htmlcom.sub(r"", t)
        t = self.rx_neg_apostrophe.sub(r"\1\2\3", t)
        t = self.rx_email.sub(r"", t)
        t = self.rx_web.sub(r"", t)
        t = self.rx_numbers.sub(r"", t)

        return t

    def convert_bad_number_representation_to_float(self, s: str) -> float:
        """Accept a string that represents a number (poorly), convert to float
        Issues:
                Currently limited to k and M.
                Hard to catch all usecases so returns nan on fail
                Yields % 100 too high
        Corrects a multitide of sins e.g:
            '1M' -> 1000000.
            '1.4m' -> 1400000.
            '25K' -> 25000.
            '81.12k' -> 81120.
            '3,000.82' -> 3000.82
            '$3,000.82' -> 3000.82
            '3.71%' -> 3.71

        ts = ['1M', '1.4m', '25K', '81.12k', '400', '9,000.23', '$35,000.82', '3.71%']
        for t in ts:
            print(t, convert_bad_number_representation_to_float(t))
        """
        r = np.nan
        s0 = self.rx_number_junk.sub(r"", str(s).strip().lower())
        gm = self.rx_num_m.match(s0)
        gk = self.rx_num_k.match(s0)
        gn = self.rx_num.match(s0)

        if gm is not None:
            mill = gm.capturesdict()["mill"]
            mill = mill[0] if len(mill) > 0 else "0"
            r = np.float64(f"{mill}") * 1e6

        elif gk is not None:
            thou = gk.capturesdict()["thou"]
            thou = thou[0] if len(thou) > 0 else "0"
            r = np.float64(f"{thou}") * 1e3

        elif gn is not None:
            whol = gn.capturesdict()["whol"]
            whol = whol[0] if len(whol) > 0 else ""
            frac = gn.capturesdict()["frac"]
            frac = frac[0] if len(frac) > 0 else ".0"
            r = np.float64(f"{whol}") + np.float64(f"{frac}")

        return r



# Copyright 2024 Oreum Industries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Core tools for use on projects by Oreum Industries"""

import logging

__version__ = "0.10.6"

# logger goes to null handler by default
# packages that import oreum_core can override this and direct elsewhere
# e.g.
#   logger = logging.getLogger('oreum_core')
#   logger.addHandler(logging.FileHandler('log.log', mode='a', encoding='utf-8'))
logging.getLogger("oreum_core").addHandler(logging.NullHandler())



# Copyright 2024 Oreum Industries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# utils.file_io.py
"""Common File IO utils"""

from pathlib import Path

from .snakey_lowercaser import SnakeyLowercaser

__all__ = ["BaseFileIO", "check_fqns_exist"]


class BaseFileIO:
    """Base handler to read/write files in a predictable way
    NOTE
    + This class is to be inherited e.g. `super().__init__(*args, **kwargs)`
    + Checks for existence of files for reading and dirs for writing
    + Allows a rootdir / rootpath as we often use in R&D Notebooks
    """

    def __init__(self, rootdir: Path = None, **kwargs):
        """Allow set a root path for convenience in Notebooks
        e.g. rootdir = DIR_MODELS_A = ['data', 'models', 'a']
        If used, then read/write will prepend this root to their input fqns
        """
        if rootdir is None:
            self.rootdir = Path().cwd()
        else:
            if not rootdir.is_dir():
                raise FileNotFoundError(
                    f"Required dir does not exist {str(self.rootdir.resolve())}"
                )
            else:
                self.rootdir = rootdir
        allow_hyphen = kwargs.get("allow_hyphen", False)
        self.snl = SnakeyLowercaser(allow_hyphen=allow_hyphen)

    def get_path_read(self, fn: str) -> Path:
        """Create and test fqn file existence for read"""
        fqn = self.rootdir.joinpath(fn)
        if not fqn.exists():
            raise FileNotFoundError(
                f"Required file does not exist {str(fqn.resolve())}"
            )
        return fqn

    def get_path_write(self, fn: str) -> Path:
        """Create and test dir existence for write, return fqn
        Ensure the passed fn is snl.cleaned"""
        fqn = self.rootdir.joinpath(fn)
        dr = Path(*fqn.parts[:-1])
        if not dr.is_dir():
            raise FileNotFoundError(f"Required dir does not exist {str(dr.resolve())}")
        return fqn


def check_fqns_exist(fqns: dict[str:Path]) -> bool:
    """Basic checks files required are present"""
    for _, path in fqns.items():
        if not path.resolve().exists():
            raise FileNotFoundError(
                f"Required file does not exist {str(path.resolve())}"
            )
    return True



# Copyright 2024 Oreum Industries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# curate.data_io.py
"""Data File Handling"""

import csv
import json
import logging
import subprocess
from pathlib import Path

import dask.dataframe as dd
import pandas as pd

from ..utils.file_io import BaseFileIO

__all__ = [
    "DaskParquetIO",
    "PandasCSVIO",
    "PandasExcelIO",
    "PandasParquetIO",
    "SimpleStringIO",
    "copy_csv2md",
]

_log = logging.getLogger(__name__)


class DaskParquetIO(BaseFileIO):
    """Simple helper class to read/write dask dataframes to parquet, including
    path and extension checking.
    """

    def __init__(self, *args, **kwargs):
        """Inherit super"""
        super().__init__(*args, **kwargs)

    def read(self, fn: str, *args, **kwargs) -> dd.DataFrame:
        """Read parquet fn from rootdir, pass args kwargs to dask.read_parquet"""
        fn = Path(fn).with_suffix(".parquet")
        fqn = self.get_path_read(fn)
        _log.info(f"Read from {str(fqn.resolve())}")
        return dd.read_parquet(path=fqn, *args, **kwargs)

    def write(self, ddf: dd.DataFrame, fn: str, *args, **kwargs) -> Path:
        """Accept dask DataFrame and fn e.g. `df.parquet`, write to fqn"""
        raise NotImplementedError
        # fqn = self.get_path_write(Path(self.snl.clean(fn)).with_suffix(".parquet"))
        # ddf.to_parquet(path=fqn, *args, **kwargs)
        # _log.info(f"Written to {str(fqn.resolve())}")
        # return fqn


class PandasCSVIO(BaseFileIO):
    """Simple helper class to read/write pandas to csv with consistency,
    including path and extension checking.
    """

    def __init__(self, *args, **kwargs):
        """Inherit super"""
        super().__init__(*args, **kwargs)

    def read(self, fn: str, *args, **kwargs) -> pd.DataFrame:
        """Read csv fn from rootdir, pass args kwargs to pd.read_csv"""
        fn = Path(fn).with_suffix(".csv")
        fqn = self.get_path_read(fn)
        _log.info(f"Read from {str(fqn.resolve())}")
        return pd.read_csv(fqn, *args, **kwargs)

    def write(self, df: pd.DataFrame, fn: str, *args, **kwargs) -> str:
        """Accept pandas DataFrame and fn e.g. `df`, write to fn.csv
        Consider using kwarg: float_format='%.3f'
        """
        fqn = self.get_path_write(Path(self.snl.clean(fn)).with_suffix(".csv"))
        kws = kwargs.copy()
        kws.update(quoting=csv.QUOTE_NONNUMERIC)
        if (len(df.index.names) == 1) & (df.index.names[0] is None):
            kws.update(index_label="rowid")
        df.to_csv(fqn, *args, **kws)
        _log.info(f"Written to {str(fqn.resolve())}")
        return fqn


class PandasExcelIO(BaseFileIO):
    """Helper class to read/write pandas to excel xlsx, using xlsxwriter.
    Includes path and extension checking.
    Write a single sheet (fire & forget), or write multiple sheets using a
    returned writer object: see eda.eda_io.output_data_dict for an example
    """

    def __init__(self, *args, **kwargs):
        """Inherit super"""
        super().__init__(*args, **kwargs)

    def read(self, fn: str, *args, **kwargs) -> pd.DataFrame:
        """Read excel fn from rootdir, pass args kwargs to pd.read_excel"""
        fn = Path(fn).with_suffix(".xlsx")
        fqn = self.get_path_read(fn)
        _log.info(f"Read from {str(fqn.resolve())}")
        return pd.read_excel(fqn, *args, **kwargs)

    def write(self, df: pd.DataFrame, fn: str, *args, **kwargs) -> Path:
        """Accept pandas DataFrame and fn e.g. `df.xlsx`, write to fqn."""
        fqn = self.get_path_write(Path(self.snl.clean(fn)).with_suffix(".xlsx"))
        writer = pd.ExcelWriter(fqn, engine="xlsxwriter")
        df.to_excel(writer, *args, **kwargs)
        writer.close()
        _log.info(f"Written to {str(fqn.resolve())}")
        return fqn

    def writer_open(self, fn: str, *args, **kwargs) -> None:
        """Starts a writer workflow to allow advanced users to write multiple sheets"""
        self.fqn = self.get_path_write(Path(self.snl.clean(fn)).with_suffix(".xlsx"))
        self.writer = pd.ExcelWriter(str(self.fqn), engine="xlsxwriter")
        _log.info(f"Opened writer workflow for {str(self.fqn.resolve())}")

    def writer_write(self, df: pd.DataFrame, *args, **kwargs) -> None:
        """Write pandas DataFrame to existing writer object"""
        if hasattr(self, "writer"):
            df.to_excel(self.writer, *args, **kwargs)
            _log.info(
                f"Written as part of writer workflow to {str(self.fqn.resolve())}"
            )

    def writer_close(self) -> Path:
        """Close existing writer object"""
        if hasattr(self, "writer"):
            self.writer.close()
            _log.info(f"Closed writer workflow {str(self.fqn.resolve())}")
            fqn = self.fqn
            del self.fqn
            del self.writer
            return fqn


class PandasParquetIO(BaseFileIO):
    """Simple helper class to read/write pandas to parquet, including path and
    extension checking.
    """

    def __init__(self, *args, **kwargs):
        """Inherit super"""
        super().__init__(*args, **kwargs)

    def read(self, fn: str, *args, **kwargs) -> pd.DataFrame:
        """Read parquet fn from rootdir, pass args kwargs to pd.read_parquet"""
        fn = Path(fn).with_suffix(".parquet")
        fqn = self.get_path_read(fn)
        _log.info(f"Read from {str(fqn.resolve())}")
        return pd.read_parquet(path=fqn, *args, **kwargs)

    def write(self, df: pd.DataFrame, fn: str, *args, **kwargs) -> Path:
        """Accept pandas DataFrame and fn e.g. `df.parquet`, write to fqn"""
        fqn = self.get_path_write(Path(self.snl.clean(fn)).with_suffix(".parquet"))
        df.to_parquet(path=fqn, *args, **kwargs)
        _log.info(f"Written to {str(fqn.resolve())}")
        return fqn


class SimpleStringIO(BaseFileIO):
    """Helper class to read/write stringlike objects to txt or json files
    Set kind to
        + 'txt' to read/write strings <-> text file
        + 'json' to read/write dicts <-> json file
    """

    def __init__(self, kind: str = "txt", *args, **kwargs):
        """Init for txt and json only"""
        super().__init__(*args, **kwargs)
        assert kind in set(["txt", "json"]), "kind must be in {'txt', 'json'}"
        self.kind = kind

    def read(self, fn: str) -> str:
        """Read a file from fn according to kind of this object"""
        fqn = self.get_path_read(fn)
        with open(str(fqn), "r") as f:
            s = f.read().rstrip("\n")
            f.close()
        _log.info(f"Read text from {str(fqn.resolve())}")
        if self.kind == "json":
            s = json.loads(s)
        return s

    def write(self, s: str, fn: str) -> str:
        fqn = self.get_path_write(Path(self.snl.clean(fn)).with_suffix(f".{self.kind}"))
        if self.kind == "json":
            s = json.dumps(s)
        with open(str(fqn), "w") as f:
            f.write(f"{s}\n")
            f.close()
        _log.info(f"Written to {str(fqn.resolve())}")
        return fqn


def copy_csv2md(fn: str) -> str:
    """Convenience to copy csv 'path/x.csv' to markdown 'path/x.md'"""
    fileio = BaseFileIO()
    fqn = fileio.get_path_read(fn)
    r = subprocess.run(["csv2md", f"{fqn}"], capture_output=True)
    fn_out = f"{fn[:-3] + 'md'}"
    fqn_out = fileio.get_path_write(fn_out)
    with open(fqn_out, "wb") as f:
        f.write(r.stdout)
        f.close()
    _log.info(f"Written to {str(fqn_out.resolve())}")
    return fqn_out



# Copyright 2024 Oreum Industries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# model.distributions.py
"""Handful of additional useful distributions / transforms"""

import numpy as np
import pymc as pm
import pytensor.tensor as pt
from pymc.distributions.dist_math import check_icdf_parameters, check_icdf_value
from scipy import stats

__all__ = ["sanity_check_lognorm", "normal_icdf", "lognormal_icdf", "mv_dist"]

# NOTE hack to clip values away from {0, 1} for invcdfs
# Whilst value = {0, 1} is theoretically allowed, it seems to cause a
# numeric computational issue somewhere in pt.erfcinv which throws infs.
# This screws up the downstream, so clip slightly away from {0, 1}
CLIP = 1e-12  # NOTE issues at 1e-15? and definitely 1e-18 too small


def sanity_check_lognorm(mu: float = 0.0, sigma: float = 1.0) -> None:
    """Sanity checker to confirm parameterisation of lognorm dists
    between scipy and pymc"""
    n = 1000
    x = np.linspace(0, 100, n)
    fd = stats.lognorm(scale=np.exp(mu), s=sigma)
    y_scipy = fd.pdf(x)
    rv = pm.LogNormal.dist(mu=mu, sigma=sigma)
    y_pymc = np.exp(pm.logp(rv, x).eval())
    assert sum(np.isclose(y_scipy, y_pymc)) == n


def normal_icdf(
    x: pt.TensorVariable, mu: pt.TensorVariable = 0.0, sigma: pt.TensorVariable = 1.0
) -> pt.TensorVariable:
    """Normal iCDF aka InverseCDF aka PPF aka Probit. Default mean-center 1sd
    NOTE:
    + Modified from pymc.distributions.continuous.Normal.icdf
    + Slow implementation of erfcinv not in C
    + See also https://stackoverflow.com/questions/60472139/computing-the-inverse-of-the-complementary-error-function-erfcinv-in-c
    + Hack to clip values away from edges [0., 1.]:
        Whilst value = {0, 1} is theoretically allowed, it seems to cause a
        numeric computational issue somewhere in pt.erfcinv which throws infs.
        This screws up the downstream, so clip slightly away from edges [0, 1]
    + Used in oreum_copula.src.copula
        NOTE: Possibly after pymc > 5.5 will change to use
        y_cop_u_rv = pm.Normal.dist(mu=0., sigma=1.)
        pm.icdf(y_cop_u_rv, pt.stack([y_m1u, y_m2u], axis=1)),
    """
    x = pt.clip(x, CLIP, 1 - CLIP)
    r = check_icdf_value(mu - sigma * pt.sqrt(2.0) * pt.erfcinv(2.0 * x), x)
    return check_icdf_parameters(r, sigma > 0.0, msg="sigma > 0")


def lognormal_icdf(
    x: pt.TensorVariable, mu: pt.TensorVariable = 0.0, sigma: pt.TensorVariable = 1.0
) -> pt.TensorVariable:
    """LogNormal icdf, defaulted to mean-centered, 1sd
    NOTE:
    + Modified from pymc.distributions.continuous.LogNormal.icdf in pymc > v5.5
      so that we can use normal_icdf above which has CLIP protection.
    + Consider replacing with pm.LogNormal.icdf(u_d, mu, s),
    """
    return pt.exp(normal_icdf(x=x, mu=mu, sigma=sigma))


def mv_dist(
    chol: pt.TensorVariable,
    n: int,
    size: pt.TensorVariable,  # required by CustomDist, but we ignore it for clarity
) -> pt.TensorVariable:
    """Hack to wrap a conventional MvNormal inside a CustomDist
    So that we can hack it to use an observed that contains a FreeRV"""
    return pm.MvNormal.dist(mu=pt.zeros(2), chol=chol, shape=(n, 2))
